# Copyright 2014 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.import mock

import mock
import six

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.io import ioutils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi import wintypes


class IOUtilsTestCase(test_base.BaseTestCase):

    _autospec_classes = [
        ioutils.win32utils.Win32Utils,
    ]

    def setUp(self):
        super(IOUtilsTestCase, self).setUp()
        self._setup_lib_mocks()

        self._ioutils = ioutils.IOUtils()

        self._mock_run = self._ioutils._win32_utils.run_and_check_output
        self._run_args = dict(kernel32_lib_func=True,
                              failure_exc=exceptions.Win32IOException,
                              eventlet_nonblocking_mode=False)

        self.addCleanup(mock.patch.stopall)

    def _setup_lib_mocks(self):
        self._ctypes = mock.Mock()
        # This is used in order to easily make assertions on the variables
        # passed by reference.
        self._ctypes.byref = lambda x: (x, "byref")
        self._ctypes.c_wchar_p = lambda x: (x, "c_wchar_p")

        mock.patch.multiple(ioutils,
                            ctypes=self._ctypes, kernel32=mock.DEFAULT,
                            create=True).start()

    def test_run_and_check_output(self):
        ret_val = self._ioutils._run_and_check_output(
            mock.sentinel.func, mock.sentinel.arg)

        self._mock_run.assert_called_once_with(mock.sentinel.func,
                                               mock.sentinel.arg,
                                               **self._run_args)
        self.assertEqual(self._mock_run.return_value, ret_val)

    def test_wait_named_pipe(self):
        fake_timeout_s = 10
        self._ioutils.wait_named_pipe(mock.sentinel.pipe_name,
                                      timeout=fake_timeout_s)

        self._mock_run.assert_called_once_with(
            ioutils.kernel32.WaitNamedPipeW,
            self._ctypes.c_wchar_p(mock.sentinel.pipe_name),
            fake_timeout_s * 1000,
            **self._run_args)

    def test_open(self):
        handle = self._ioutils.open(mock.sentinel.path,
                                    mock.sentinel.access,
                                    mock.sentinel.share_mode,
                                    mock.sentinel.create_disposition,
                                    mock.sentinel.flags)

        self._mock_run.assert_called_once_with(
            ioutils.kernel32.CreateFileW,
            self._ctypes.c_wchar_p(mock.sentinel.path),
            mock.sentinel.access,
            mock.sentinel.share_mode,
            None,
            mock.sentinel.create_disposition,
            mock.sentinel.flags,
            None,
            error_ret_vals=[w_const.INVALID_HANDLE_VALUE],
            **self._run_args)
        self.assertEqual(self._mock_run.return_value, handle)

    def test_cancel_io(self):
        self._ioutils.cancel_io(mock.sentinel.handle,
                                mock.sentinel.overlapped_struct,
                                ignore_invalid_handle=True)

        expected_ignored_err_codes = [w_const.ERROR_NOT_FOUND,
                                      w_const.ERROR_INVALID_HANDLE]

        self._mock_run.assert_called_once_with(
            ioutils.kernel32.CancelIoEx,
            mock.sentinel.handle,
            self._ctypes.byref(mock.sentinel.overlapped_struct),
            ignored_error_codes=expected_ignored_err_codes,
            **self._run_args)

    def test_close_handle(self):
        self._ioutils.close_handle(mock.sentinel.handle)

        self._mock_run.assert_called_once_with(ioutils.kernel32.CloseHandle,
                                               mock.sentinel.handle,
                                               **self._run_args)

    def test_wait_io_completion(self):
        self._ioutils._wait_io_completion(mock.sentinel.event)

        self._mock_run.assert_called_once_with(
            ioutils.kernel32.WaitForSingleObjectEx,
            mock.sentinel.event,
            ioutils.WAIT_INFINITE_TIMEOUT,
            True,
            error_ret_vals=[w_const.WAIT_FAILED],
            **self._run_args)

    def test_set_event(self):
        self._ioutils.set_event(mock.sentinel.event)

        self._mock_run.assert_called_once_with(ioutils.kernel32.SetEvent,
                                               mock.sentinel.event,
                                               **self._run_args)

    def test_reset_event(self):
        self._ioutils._reset_event(mock.sentinel.event)

        self._mock_run.assert_called_once_with(ioutils.kernel32.ResetEvent,
                                               mock.sentinel.event,
                                               **self._run_args)

    def test_create_event(self):
        event = self._ioutils._create_event(mock.sentinel.event_attributes,
                                            mock.sentinel.manual_reset,
                                            mock.sentinel.initial_state,
                                            mock.sentinel.name)

        self._mock_run.assert_called_once_with(ioutils.kernel32.CreateEventW,
                                               mock.sentinel.event_attributes,
                                               mock.sentinel.manual_reset,
                                               mock.sentinel.initial_state,
                                               mock.sentinel.name,
                                               error_ret_vals=[None],
                                               **self._run_args)
        self.assertEqual(self._mock_run.return_value, event)

    @mock.patch.object(wintypes, 'LPOVERLAPPED', create=True)
    @mock.patch.object(wintypes, 'LPOVERLAPPED_COMPLETION_ROUTINE',
                       lambda x: x, create=True)
    @mock.patch.object(ioutils.IOUtils, 'set_event')
    def test_get_completion_routine(self, mock_set_event,
                                    mock_LPOVERLAPPED):
        mock_callback = mock.Mock()

        compl_routine = self._ioutils.get_completion_routine(mock_callback)
        compl_routine(mock.sentinel.error_code,
                      mock.sentinel.num_bytes,
                      mock.sentinel.lpOverLapped)

        self._ctypes.cast.assert_called_once_with(mock.sentinel.lpOverLapped,
                                                  wintypes.LPOVERLAPPED)
        mock_overlapped_struct = self._ctypes.cast.return_value.contents
        mock_set_event.assert_called_once_with(mock_overlapped_struct.hEvent)
        mock_callback.assert_called_once_with(mock.sentinel.num_bytes)

    @mock.patch.object(wintypes, 'OVERLAPPED', create=True)
    @mock.patch.object(ioutils.IOUtils, '_create_event')
    def test_get_new_overlapped_structure(self, mock_create_event,
                                          mock_OVERLAPPED):
        overlapped_struct = self._ioutils.get_new_overlapped_structure()

        self.assertEqual(mock_OVERLAPPED.return_value, overlapped_struct)
        self.assertEqual(mock_create_event.return_value,
                         overlapped_struct.hEvent)

    @mock.patch.object(ioutils.IOUtils, '_reset_event')
    @mock.patch.object(ioutils.IOUtils, '_wait_io_completion')
    def test_read(self, mock_wait_io_completion, mock_reset_event):
        mock_overlapped_struct = mock.Mock()
        mock_event = mock_overlapped_struct.hEvent
        self._ioutils.read(mock.sentinel.handle, mock.sentinel.buff,
                           mock.sentinel.num_bytes,
                           mock_overlapped_struct,
                           mock.sentinel.compl_routine)

        mock_reset_event.assert_called_once_with(mock_event)
        self._mock_run.assert_called_once_with(ioutils.kernel32.ReadFileEx,
                                               mock.sentinel.handle,
                                               mock.sentinel.buff,
                                               mock.sentinel.num_bytes,
                                               self._ctypes.byref(
                                                   mock_overlapped_struct),
                                               mock.sentinel.compl_routine,
                                               **self._run_args)
        mock_wait_io_completion.assert_called_once_with(mock_event)

    @mock.patch.object(ioutils.IOUtils, '_reset_event')
    @mock.patch.object(ioutils.IOUtils, '_wait_io_completion')
    def test_write(self, mock_wait_io_completion, mock_reset_event):
        mock_overlapped_struct = mock.Mock()
        mock_event = mock_overlapped_struct.hEvent
        self._ioutils.write(mock.sentinel.handle, mock.sentinel.buff,
                            mock.sentinel.num_bytes,
                            mock_overlapped_struct,
                            mock.sentinel.compl_routine)

        mock_reset_event.assert_called_once_with(mock_event)
        self._mock_run.assert_called_once_with(ioutils.kernel32.WriteFileEx,
                                               mock.sentinel.handle,
                                               mock.sentinel.buff,
                                               mock.sentinel.num_bytes,
                                               self._ctypes.byref(
                                                   mock_overlapped_struct),
                                               mock.sentinel.compl_routine,
                                               **self._run_args)
        mock_wait_io_completion.assert_called_once_with(mock_event)

    def test_buffer_ops(self):
        mock.patch.stopall()

        fake_data = 'fake data'

        buff = self._ioutils.get_buffer(len(fake_data), data=fake_data)
        buff_data = self._ioutils.get_buffer_data(buff, len(fake_data))

        self.assertEqual(six.b(fake_data), buff_data)


class IOQueueTestCase(test_base.BaseTestCase):
    def setUp(self):
        super(IOQueueTestCase, self).setUp()

        self._mock_queue = mock.Mock()
        queue_patcher = mock.patch.object(ioutils.Queue, 'Queue',
                                          new=self._mock_queue)
        queue_patcher.start()
        self.addCleanup(queue_patcher.stop)

        self._mock_client_connected = mock.Mock()
        self._ioqueue = ioutils.IOQueue(self._mock_client_connected)

    def test_get(self):
        self._mock_client_connected.isSet.return_value = True
        self._mock_queue.get.return_value = mock.sentinel.item

        queue_item = self._ioqueue.get(timeout=mock.sentinel.timeout)

        self._mock_queue.get.assert_called_once_with(
            self._ioqueue, timeout=mock.sentinel.timeout)
        self.assertEqual(mock.sentinel.item, queue_item)

    def _test_get_timeout(self, continue_on_timeout=True):
        self._mock_client_connected.isSet.side_effect = [True, True, False]
        self._mock_queue.get.side_effect = ioutils.Queue.Empty

        queue_item = self._ioqueue.get(timeout=mock.sentinel.timeout,
                                       continue_on_timeout=continue_on_timeout)

        expected_calls_number = 2 if continue_on_timeout else 1
        self._mock_queue.get.assert_has_calls(
            [mock.call(self._ioqueue, timeout=mock.sentinel.timeout)] *
            expected_calls_number)
        self.assertIsNone(queue_item)

    def test_get_continue_on_timeout(self):
        # Test that the queue blocks as long
        # as the client connected event is set.
        self._test_get_timeout()

    def test_get_break_on_timeout(self):
        self._test_get_timeout(continue_on_timeout=False)

    def test_put(self):
        self._mock_client_connected.isSet.side_effect = [True, True, False]
        self._mock_queue.put.side_effect = ioutils.Queue.Full

        self._ioqueue.put(mock.sentinel.item,
                          timeout=mock.sentinel.timeout)

        self._mock_queue.put.assert_has_calls(
            [mock.call(self._ioqueue, mock.sentinel.item,
                       timeout=mock.sentinel.timeout)] * 2)

    @mock.patch.object(ioutils.IOQueue, 'get')
    def _test_get_burst(self, mock_get,
                        exceeded_max_size=False):
        fake_data = 'fake_data'

        mock_get.side_effect = [fake_data, fake_data, None]

        if exceeded_max_size:
            max_size = 0
        else:
            max_size = constants.SERIAL_CONSOLE_BUFFER_SIZE

        ret_val = self._ioqueue.get_burst(
            timeout=mock.sentinel.timeout,
            burst_timeout=mock.sentinel.burst_timeout,
            max_size=max_size)

        expected_calls = [mock.call(timeout=mock.sentinel.timeout)]
        expected_ret_val = fake_data

        if not exceeded_max_size:
            expected_calls.append(
                mock.call(timeout=mock.sentinel.burst_timeout,
                          continue_on_timeout=False))
            expected_ret_val += fake_data

        mock_get.assert_has_calls(expected_calls)
        self.assertEqual(expected_ret_val, ret_val)

    def test_get_burst(self):
        self._test_get_burst()

    def test_get_burst_exceeded_size(self):
        self._test_get_burst(exceeded_max_size=True)



# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the Hyper-V NVGRE support.
"""

import mock

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.network import nvgreutils


class TestNvgreUtils(test_base.OsWinBaseTestCase):

    _FAKE_RDID = 'fake_rdid'
    _FAKE_NETWORK_NAME = 'fake_network_name'
    _FAKE_VSID = 9001
    _FAKE_DEST_PREFIX = 'fake_dest_prefix'
    _FAKE_GW_BAD = '10.0.0.1'
    _FAKE_GW = '10.0.0.2'

    def setUp(self):
        super(TestNvgreUtils, self).setUp()
        self.utils = nvgreutils.NvgreUtils()
        self.utils._utils = mock.MagicMock()
        self.utils._scimv2 = mock.MagicMock()

    def _create_mock_binding(self):
        binding = mock.MagicMock()
        binding.BindName = self.utils._WNV_BIND_NAME
        binding.Name = mock.sentinel.fake_network

        net_binds = self.utils._scimv2.MSFT_NetAdapterBindingSettingData
        net_binds.return_value = [binding]
        return binding

    @mock.patch.object(nvgreutils.NvgreUtils, 'get_network_iface_ip')
    @mock.patch.object(nvgreutils.NvgreUtils, '_get_network_iface_index')
    def test_create_provider_address(self, mock_get_iface_index,
                                     mock_get_iface_ip):
        mock_get_iface_index.return_value = mock.sentinel.iface_index
        mock_get_iface_ip.return_value = (mock.sentinel.iface_ip,
                                          mock.sentinel.prefix_len)

        provider_addr = mock.MagicMock()
        scimv2 = self.utils._scimv2
        obj_class = scimv2.MSFT_NetVirtualizationProviderAddressSettingData
        obj_class.return_value = [provider_addr]

        self.utils.create_provider_address(mock.sentinel.fake_network,
                                           mock.sentinel.fake_vlan_id)

        self.assertTrue(provider_addr.Delete_.called)
        obj_class.new.assert_called_once_with(
            ProviderAddress=mock.sentinel.iface_ip,
            VlanID=mock.sentinel.fake_vlan_id,
            InterfaceIndex=mock.sentinel.iface_index,
            PrefixLength=mock.sentinel.prefix_len)

    @mock.patch.object(nvgreutils.NvgreUtils, 'get_network_iface_ip')
    @mock.patch.object(nvgreutils.NvgreUtils, '_get_network_iface_index')
    def test_create_provider_address_exc(self, mock_get_iface_index,
                                         mock_get_iface_ip):
        mock_get_iface_ip.return_value = (None, None)

        self.assertRaises(exceptions.NotFound,
                          self.utils.create_provider_address,
                          mock.sentinel.fake_network,
                          mock.sentinel.fake_vlan_id)

    @mock.patch.object(nvgreutils.NvgreUtils, 'get_network_iface_ip')
    @mock.patch.object(nvgreutils.NvgreUtils, '_get_network_iface_index')
    def test_create_provider_address_exists(self, mock_get_iface_index,
                                            mock_get_iface_ip):
        mock_get_iface_index.return_value = mock.sentinel.iface_index
        mock_get_iface_ip.return_value = (mock.sentinel.iface_ip,
                                          mock.sentinel.prefix_len)

        provider_addr = mock.MagicMock(
            VlanID=mock.sentinel.fake_vlan_id,
            InterfaceIndex=mock.sentinel.iface_index)
        scimv2 = self.utils._scimv2
        obj_class = scimv2.MSFT_NetVirtualizationProviderAddressSettingData
        obj_class.return_value = [provider_addr]

        self.utils.create_provider_address(mock.sentinel.fake_network,
                                           mock.sentinel.fake_vlan_id)

        self.assertFalse(obj_class.new.called)

    @mock.patch.object(nvgreutils.NvgreUtils, '_get_network_iface_index')
    def test_create_provider_route(self, mock_get_iface_index):
        mock_get_iface_index.return_value = mock.sentinel.iface_index
        self.utils._scimv2.MSFT_NetVirtualizationProviderRouteSettingData = (
            mock.MagicMock(return_value=[]))

        self.utils.create_provider_route(mock.sentinel.fake_network)

        scimv2 = self.utils._scimv2
        obj_class = scimv2.MSFT_NetVirtualizationProviderRouteSettingData
        obj_class.new.assert_called_once_with(
            InterfaceIndex=mock.sentinel.iface_index,
            DestinationPrefix='%s/0' % constants.IPV4_DEFAULT,
            NextHop=constants.IPV4_DEFAULT)

    @mock.patch.object(nvgreutils.NvgreUtils, '_get_network_iface_index')
    def test_create_provider_route_none(self, mock_get_iface_index):
        mock_get_iface_index.return_value = None

        self.utils.create_provider_route(mock.sentinel.fake_network)
        scimv2 = self.utils._scimv2
        self.assertFalse(
            scimv2.MSFT_NetVirtualizationProviderRouteSettingData.new.called)

    @mock.patch.object(nvgreutils.NvgreUtils, '_get_network_iface_index')
    def test_create_provider_route_exists(self, mock_get_iface_index):
        mock_get_iface_index.return_value = mock.sentinel.iface_index
        self.utils._scimv2.MSFT_NetVirtualizationProviderRouteSettingData = (
            mock.MagicMock(return_value=[mock.MagicMock()]))

        self.utils.create_provider_route(mock.sentinel.fake_network)

        scimv2 = self.utils._scimv2
        self.assertFalse(
            scimv2.MSFT_NetVirtualizationProviderRouteSettingData.new.called)

    def test_clear_customer_routes(self):
        cls = self.utils._scimv2.MSFT_NetVirtualizationCustomerRouteSettingData
        route = mock.MagicMock()
        cls.return_value = [route]

        self.utils.clear_customer_routes(mock.sentinel.vsid)

        cls.assert_called_once_with(VirtualSubnetID=mock.sentinel.vsid)
        route.Delete_.assert_called_once_with()

    def test_create_customer_route(self):
        self.utils.create_customer_route(
            mock.sentinel.fake_vsid, mock.sentinel.dest_prefix,
            mock.sentinel.next_hop, self._FAKE_RDID)

        scimv2 = self.utils._scimv2
        obj_class = scimv2.MSFT_NetVirtualizationCustomerRouteSettingData
        obj_class.new.assert_called_once_with(
            VirtualSubnetID=mock.sentinel.fake_vsid,
            DestinationPrefix=mock.sentinel.dest_prefix,
            NextHop=mock.sentinel.next_hop,
            Metric=255,
            RoutingDomainID='{%s}' % self._FAKE_RDID)

    def _check_create_lookup_record(self, customer_addr, expected_type):
        lookup = mock.MagicMock()
        scimv2 = self.utils._scimv2
        obj_class = scimv2.MSFT_NetVirtualizationLookupRecordSettingData
        obj_class.return_value = [lookup]

        self.utils.create_lookup_record(mock.sentinel.provider_addr,
                                        customer_addr,
                                        mock.sentinel.mac_addr,
                                        mock.sentinel.fake_vsid)

        self.assertTrue(lookup.Delete_.called)
        obj_class.new.assert_called_once_with(
            VirtualSubnetID=mock.sentinel.fake_vsid,
            Rule=self.utils._TRANSLATE_ENCAP,
            Type=expected_type,
            MACAddress=mock.sentinel.mac_addr,
            CustomerAddress=customer_addr,
            ProviderAddress=mock.sentinel.provider_addr)

    def test_create_lookup_record_l2_only(self):
        self._check_create_lookup_record(
            constants.IPV4_DEFAULT,
            self.utils._LOOKUP_RECORD_TYPE_L2_ONLY)

    def test_create_lookup_record_static(self):
        self._check_create_lookup_record(
            mock.sentinel.customer_addr,
            self.utils._LOOKUP_RECORD_TYPE_STATIC)

    def test_create_lookup_record_exists(self):
        lookup = mock.MagicMock(VirtualSubnetID=mock.sentinel.fake_vsid,
                                ProviderAddress=mock.sentinel.provider_addr,
                                CustomerAddress=mock.sentinel.customer_addr,
                                MACAddress=mock.sentinel.mac_addr)
        scimv2 = self.utils._scimv2
        obj_class = scimv2.MSFT_NetVirtualizationLookupRecordSettingData
        obj_class.return_value = [lookup]

        self.utils.create_lookup_record(mock.sentinel.provider_addr,
                                        mock.sentinel.customer_addr,
                                        mock.sentinel.mac_addr,
                                        mock.sentinel.fake_vsid)
        self.assertFalse(obj_class.new.called)

    def test_get_network_iface_index_cached(self):
        self.utils._net_if_indexes[mock.sentinel.fake_network] = (
            mock.sentinel.iface_index)

        index = self.utils._get_network_iface_index(mock.sentinel.fake_network)

        self.assertEqual(mock.sentinel.iface_index, index)
        self.assertFalse(self.utils._scimv2.MSFT_NetAdapter.called)

    def test_get_network_iface_index_not_found(self):
        self.utils._scimv2.MSFT_NetAdapter.return_value = []
        self.assertRaises(exceptions.NotFound,
                          self.utils._get_network_iface_index,
                          mock.sentinel.network_name)

    def test_get_network_iface_index(self):
        fake_network = mock.MagicMock(InterfaceIndex=mock.sentinel.iface_index)
        self.utils._scimv2.MSFT_NetAdapter.return_value = [fake_network]
        description = (
            self.utils._utils.get_vswitch_external_network_name.return_value)

        index = self.utils._get_network_iface_index(mock.sentinel.fake_network)

        self.assertEqual(mock.sentinel.iface_index, index)
        self.assertIn(mock.sentinel.fake_network, self.utils._net_if_indexes)
        self.utils._scimv2.MSFT_NetAdapter.assert_called_once_with(
            InterfaceDescription=description)

    @mock.patch.object(nvgreutils.NvgreUtils, '_get_network_ifaces_by_name')
    def test_get_network_iface_ip(self, mock_get_net_ifaces):
        fake_network = mock.MagicMock(
            InterfaceIndex=mock.sentinel.iface_index,
            DriverDescription=self.utils._HYPERV_VIRT_ADAPTER)
        mock_get_net_ifaces.return_value = [fake_network]

        fake_netip = mock.MagicMock(IPAddress=mock.sentinel.provider_addr,
                                    PrefixLength=mock.sentinel.prefix_len)
        self.utils._scimv2.MSFT_NetIPAddress.return_value = [fake_netip]

        pair = self.utils.get_network_iface_ip(mock.sentinel.fake_network)

        self.assertEqual(
            (mock.sentinel.provider_addr, mock.sentinel.prefix_len), pair)

    @mock.patch.object(nvgreutils.NvgreUtils, '_get_network_ifaces_by_name')
    def test_get_network_iface_ip_none(self, mock_get_net_ifaces):
        mock_get_net_ifaces.return_value = []
        pair = self.utils.get_network_iface_ip(mock.sentinel.fake_network)
        self.assertEqual((None, None), pair)



# Copyright 2015 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock

from os_win import _utils
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils import win32utils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi import wintypes


@ddt.ddt
class Win32UtilsTestCase(test_base.BaseTestCase):
    def setUp(self):
        super(Win32UtilsTestCase, self).setUp()
        self._setup_lib_mocks()

        self._win32_utils = win32utils.Win32Utils()

        self.addCleanup(mock.patch.stopall)

    def _setup_lib_mocks(self):
        self._ctypes = mock.Mock()
        # This is used in order to easily make assertions on the variables
        # passed by reference.
        self._ctypes.byref = lambda x: (x, "byref")

        self._ctypes_patcher = mock.patch.multiple(
            win32utils, ctypes=self._ctypes)
        self._ctypes_patcher.start()

        mock.patch.multiple(win32utils,
                            kernel32=mock.DEFAULT,
                            create=True).start()

    @mock.patch.object(win32utils.Win32Utils, 'get_error_message')
    @mock.patch.object(win32utils.Win32Utils, 'get_last_error')
    def _test_run_and_check_output(self, mock_get_last_err, mock_get_err_msg,
                                   ret_val=0, expected_exc=None,
                                   **kwargs):
        self._ctypes_patcher.stop()

        mock_func = mock.Mock()
        mock_func.return_value = ret_val

        if expected_exc:
            self.assertRaises(expected_exc,
                              self._win32_utils.run_and_check_output,
                              mock_func,
                              mock.sentinel.arg,
                              kwarg=mock.sentinel.kwarg,
                              **kwargs)
        else:
            actual_ret_val = self._win32_utils.run_and_check_output(
                mock_func,
                mock.sentinel.arg,
                kwarg=mock.sentinel.kwarg,
                **kwargs)
            self.assertEqual(ret_val, actual_ret_val)

        mock_func.assert_called_once_with(mock.sentinel.arg,
                                          kwarg=mock.sentinel.kwarg)

        return mock_get_last_err, mock_get_err_msg

    def test_run_and_check_output(self):
        self._test_run_and_check_output()

    def test_run_and_check_output_fail_on_nonzero_ret_val(self):
        ret_val = 1

        (mock_get_last_err,
         mock_get_err_msg) = self._test_run_and_check_output(
            ret_val=ret_val,
            expected_exc=exceptions.VHDWin32APIException,
            failure_exc=exceptions.VHDWin32APIException)

        mock_get_err_msg.assert_called_once_with(ret_val)

    def test_run_and_check_output_explicit_error_ret_vals(self):
        ret_val = 1
        error_ret_vals = [ret_val]

        (mock_get_last_err,
         mock_get_err_msg) = self._test_run_and_check_output(
            ret_val=ret_val,
            error_ret_vals=error_ret_vals,
            ret_val_is_err_code=False,
            expected_exc=exceptions.Win32Exception)

        mock_get_err_msg.assert_called_once_with(
            win32utils.ctypes.c_ulong(mock_get_last_err).value)

    def test_run_and_check_output_ignored_error(self):
        ret_val = 1
        ignored_err_codes = [ret_val]

        self._test_run_and_check_output(ret_val=ret_val,
                                        ignored_error_codes=ignored_err_codes)

    def test_run_and_check_output_kernel32_lib_func(self):
        ret_val = 0
        self._test_run_and_check_output(ret_val=ret_val,
                                        expected_exc=exceptions.Win32Exception,
                                        kernel32_lib_func=True)

    def test_run_and_check_output_with_err_msg_dict(self):
        self._ctypes_patcher.stop()

        err_code = 1
        err_msg = 'fake_err_msg'
        err_msg_dict = {err_code: err_msg}

        mock_func = mock.Mock()
        mock_func.return_value = err_code

        try:
            self._win32_utils.run_and_check_output(mock_func,
                                                   mock.sentinel.arg,
                                                   error_msg_src=err_msg_dict)
        except Exception as ex:
            self.assertIsInstance(ex, exceptions.Win32Exception)
            self.assertIn(err_msg, ex.message)

    @mock.patch.object(win32utils.Win32Utils, '_run_and_check_output')
    def test_run_and_check_output_eventlet_nb_mode_disabled(self, mock_helper):
        self._win32_utils.run_and_check_output(
            mock.sentinel.func,
            mock.sentinel.arg,
            eventlet_nonblocking_mode=False)
        mock_helper.assert_called_once_with(mock.sentinel.func,
                                            mock.sentinel.arg)

    @mock.patch.object(_utils, 'avoid_blocking_call')
    def test_run_and_check_output_eventlet_nb_mode_enabled(self, mock_helper):
        self._win32_utils.run_and_check_output(
            mock.sentinel.func,
            mock.sentinel.arg,
            eventlet_nonblocking_mode=True)
        mock_helper.assert_called_once_with(
            self._win32_utils._run_and_check_output,
            mock.sentinel.func,
            mock.sentinel.arg)

    def test_get_error_message(self):
        err_msg = self._win32_utils.get_error_message(mock.sentinel.err_code)

        fake_msg_buff = win32utils.ctypes.c_char_p.return_value

        expected_flags = (w_const.FORMAT_MESSAGE_FROM_SYSTEM |
                          w_const.FORMAT_MESSAGE_ALLOCATE_BUFFER |
                          w_const.FORMAT_MESSAGE_IGNORE_INSERTS)

        win32utils.kernel32.FormatMessageA.assert_called_once_with(
            expected_flags, None, mock.sentinel.err_code, 0,
            win32utils.ctypes.byref(fake_msg_buff), 0, None)
        self.assertEqual(fake_msg_buff.value, err_msg)

    def test_get_last_error(self):
        last_err = self._win32_utils.get_last_error()

        self.assertEqual(win32utils.kernel32.GetLastError.return_value,
                         last_err)
        win32utils.kernel32.SetLastError.assert_called_once_with(0)

    def test_hresult_to_err_code(self):
        # This could differ based on the error source.
        # Only the last 2 bytes of the hresult the error code.
        fake_file_exists_hres = -0x7ff8ffb0
        file_exists_err_code = 0x50

        ret_val = self._win32_utils.hresult_to_err_code(fake_file_exists_hres)
        self.assertEqual(file_exists_err_code, ret_val)

    @mock.patch.object(win32utils._utils, 'get_com_error_hresult')
    @mock.patch.object(win32utils.Win32Utils, 'hresult_to_err_code')
    def test_get_com_err_code(self, mock_hres_to_err_code, mock_get_hresult):
        ret_val = self._win32_utils.get_com_err_code(mock.sentinel.com_err)

        self.assertEqual(mock_hres_to_err_code.return_value, ret_val)
        mock_get_hresult.assert_called_once_with(mock.sentinel.com_err)
        mock_hres_to_err_code.assert_called_once_with(
            mock_get_hresult.return_value)

    @ddt.data(0, 1)
    @mock.patch.object(win32utils.LOG, 'exception')
    def test_local_free(self, ret_val, mock_log_exc):
        mock_localfree = win32utils.kernel32.LocalFree
        mock_localfree.return_value = ret_val

        self._win32_utils.local_free(mock.sentinel.handle)

        mock_localfree.assert_any_call(mock.sentinel.handle)
        self.assertEqual(bool(ret_val), mock_log_exc.called)

    @mock.patch.object(win32utils.Win32Utils, 'run_and_check_output')
    def test_wait_for_multiple_objects(self, mock_helper):
        fake_handles = [10, 11]

        ret_val = self._win32_utils.wait_for_multiple_objects(
            fake_handles, mock.sentinel.wait_all, mock.sentinel.milliseconds)

        mock_helper.assert_called_once_with(
            win32utils.kernel32.WaitForMultipleObjects,
            len(fake_handles),
            mock.ANY,
            mock.sentinel.wait_all,
            mock.sentinel.milliseconds,
            kernel32_lib_func=True,
            error_ret_vals=[w_const.WAIT_FAILED])
        self.assertEqual(mock_helper.return_value, ret_val)

        handles_arg = mock_helper.call_args_list[0][0][2]
        self.assertIsInstance(handles_arg,
                              wintypes.HANDLE * len(fake_handles))
        self.assertEqual(fake_handles, handles_arg[:])

    @mock.patch.object(win32utils.Win32Utils, 'run_and_check_output')
    def test_wait_for_multiple_objects_timeout(self, mock_helper):
        fake_handles = [10]
        mock_helper.return_value = w_const.ERROR_WAIT_TIMEOUT

        self.assertRaises(
            exceptions.Timeout,
            self._win32_utils.wait_for_multiple_objects,
            fake_handles, mock.sentinel.wait_all,
            mock.sentinel.milliseconds)



# Copyright 2015 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes
import inspect
from pkg_resources import parse_version
import textwrap
import time
import types

import eventlet
from eventlet import tpool
import netaddr
from oslo_concurrency import lockutils
from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import reflection
import six

from os_win import constants
from os_win import exceptions


LOG = logging.getLogger(__name__)

socket = eventlet.import_patched('socket')
synchronized = lockutils.synchronized_with_prefix('oswin-')

_WBEM_E_NOT_FOUND = 0x80041002


def execute(*cmd, **kwargs):
    """Convenience wrapper around oslo's execute() method."""
    return processutils.execute(*cmd, **kwargs)


def parse_server_string(server_str):
    """Parses the given server_string and returns a tuple of host and port.

    If it's not a combination of host part and port, the port element
    is an empty string. If the input is invalid expression, return a tuple of
    two empty strings.
    """

    try:
        # First of all, exclude pure IPv6 address (w/o port).
        if netaddr.valid_ipv6(server_str):
            return (server_str, '')

        # Next, check if this is IPv6 address with a port number combination.
        if server_str.find("]:") != -1:
            (address, port) = server_str.replace('[', '', 1).split(']:')
            return (address, port)

        # Third, check if this is a combination of an address and a port
        if server_str.find(':') == -1:
            return (server_str, '')

        # This must be a combination of an address and a port
        (address, port) = server_str.split(':')
        return (address, port)

    except (ValueError, netaddr.AddrFormatError):
        LOG.error('Invalid server_string: %s', server_str)
        return ('', '')


def get_wrapped_function(function):
    """Get the method at the bottom of a stack of decorators."""
    if not hasattr(function, '__closure__') or not function.__closure__:
        return function

    def _get_wrapped_function(function):
        if not hasattr(function, '__closure__') or not function.__closure__:
            return None

        for closure in function.__closure__:
            func = closure.cell_contents

            deeper_func = _get_wrapped_function(func)
            if deeper_func:
                return deeper_func
            elif isinstance(closure.cell_contents, types.FunctionType):
                return closure.cell_contents

    return _get_wrapped_function(function)


def retry_decorator(max_retry_count=5, timeout=None, inc_sleep_time=1,
                    max_sleep_time=1, exceptions=(), error_codes=(),
                    pass_retry_context=False):
    """Retries invoking the decorated method in case of expected exceptions.

    :param max_retry_count: The maximum number of retries performed. If 0, no
                            retry is performed. If None, there will be no limit
                            on the number of retries.
    :param timeout: The maximum time for which we'll retry invoking the method.
                    If 0 or None, there will be no time limit.
    :param inc_sleep_time: The time sleep increment used between retries.
    :param max_sleep_time: The maximum time to wait between retries.
    :param exceptions: A list of expected exceptions for which retries will be
                       performed.
    :param error_codes: A list of expected error codes. The error code is
                        retrieved from the 'error_code' exception attribute,
                        for example in case of Win32Exception. If this argument
                        is not passed, retries will be performed for any of the
                        expected exceptions.
    :param pass_retry_context: Convenient way of letting a method aware of
                               this decorator prevent a retry from being
                               performed. The decorated method must accept an
                               argument called 'retry_context', which will
                               include a dict containing the 'prevent_retry'
                               field. If this field is set, no further retries
                               will be performed.
    """

    if isinstance(error_codes, six.integer_types):
        error_codes = (error_codes, )

    def wrapper(f):
        def inner(*args, **kwargs):
            try_count = 0
            sleep_time = 0
            time_start = time.time()

            retry_context = dict(prevent_retry=False)
            if pass_retry_context:
                kwargs['retry_context'] = retry_context

            while True:
                try:
                    return f(*args, **kwargs)
                except exceptions as exc:
                    with excutils.save_and_reraise_exception() as ctxt:
                        err_code = getattr(exc, 'error_code', None)
                        expected_err_code = (err_code in error_codes or not
                                             error_codes)

                        time_elapsed = time.time() - time_start
                        time_left = (timeout - time_elapsed
                                     if timeout else 'undefined')
                        tries_left = (max_retry_count - try_count
                                      if max_retry_count is not None
                                      else 'undefined')

                        should_retry = (
                            not retry_context['prevent_retry'] and
                            expected_err_code and
                            tries_left and
                            (time_left == 'undefined' or
                             time_left > 0))
                        ctxt.reraise = not should_retry

                        if should_retry:
                            try_count += 1
                            func_name = reflection.get_callable_name(f)

                            sleep_time = min(sleep_time + inc_sleep_time,
                                             max_sleep_time)
                            if timeout:
                                sleep_time = min(sleep_time, time_left)

                            LOG.debug("Got expected exception %(exc)s while "
                                      "calling function %(func_name)s. "
                                      "Retries left: %(retries_left)s. "
                                      "Time left: %(time_left)s. "
                                      "Time elapsed: %(time_elapsed)s "
                                      "Retrying in %(sleep_time)s seconds.",
                                      dict(exc=exc,
                                           func_name=func_name,
                                           retries_left=tries_left,
                                           time_left=time_left,
                                           time_elapsed=time_elapsed,
                                           sleep_time=sleep_time))
                            time.sleep(sleep_time)
        return inner
    return wrapper


def get_ips(addr):
    addr_info = socket.getaddrinfo(addr, None, 0, 0, 0)
    # Returns IPv4 and IPv6 addresses, ordered by protocol family
    addr_info.sort()
    return [a[4][0] for a in addr_info]


def avoid_blocking_call(f, *args, **kwargs):
    """Ensures that the invoked method will not block other greenthreads.

    Performs the call in a different thread using tpool.execute when called
    from a greenthread.
    """
    # Note that eventlet.getcurrent will always return a greenlet object.
    # In case of a greenthread, the parent greenlet will always be the hub
    # loop greenlet.
    if eventlet.getcurrent().parent:
        return tpool.execute(f, *args, **kwargs)
    else:
        return f(*args, **kwargs)


def avoid_blocking_call_decorator(f):
    def wrapper(*args, **kwargs):
        return avoid_blocking_call(f, *args, **kwargs)
    return wrapper


def get_com_error_hresult(com_error):
    try:
        return ctypes.c_uint(com_error.excepinfo[5]).value
    except Exception:
        LOG.debug("Unable to retrieve COM error hresult: %s", com_error)


def _is_not_found_exc(exc):
    hresult = get_com_error_hresult(exc.com_error)
    return hresult == _WBEM_E_NOT_FOUND


def not_found_decorator(translated_exc=exceptions.NotFound):
    """Wraps x_wmi: Not Found exceptions as os_win.exceptions.NotFound."""

    def wrapper(func):
        def inner(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exceptions.x_wmi as ex:
                if _is_not_found_exc(ex):
                    LOG.debug('x_wmi: Not Found exception raised while '
                              'running %s', func.__name__)
                    raise translated_exc(message=six.text_type(ex))
                raise
        return inner
    return wrapper


def hex_str_to_byte_array(string):
    string = string.lower().replace('0x', '')
    if len(string) % 2:
        string = "0%s" % string

    return bytearray(
        [int(hex_byte, 16) for hex_byte in textwrap.wrap(string, 2)])


def byte_array_to_hex_str(byte_aray):
    return ''.join('{:02X}'.format(b) for b in byte_aray)


def required_vm_version(min_version=constants.VM_VERSION_5_0,
                        max_version=constants.VM_VERSION_254_0):
    """Ensures that the wrapped method's VM meets the version requirements.

    Some Hyper-V operations require a minimum VM version in order to succeed.
    For example, Production Checkpoints are supported on VM Versions 6.2 and
    newer.

    Clustering Hyper-V compute nodes may change the list of supported VM
    versions list and the default VM version on that host.

    :param min_version: string, the VM's minimum version required for the
        operation to succeed.
    :param max_version: string, the VM's maximum version required for the
        operation to succeed.
    :raises exceptions.InvalidVMVersion: if the VM's version does not meet the
        given requirements.
    """

    def wrapper(func):
        def inner(*args, **kwargs):
            all_args = inspect.getcallargs(func, *args, **kwargs)
            vmsettings = all_args['vmsettings']

            # NOTE(claudiub): VMs on Windows / Hyper-V Server 2012 do not have
            # a Version field, but they are 4.0.
            vm_version_str = getattr(vmsettings, 'Version', '4.0')
            vm_version = parse_version(vm_version_str)
            if (vm_version >= parse_version(min_version) and
                    vm_version <= parse_version(max_version)):
                return func(*args, **kwargs)

            raise exceptions.InvalidVMVersion(
                vm_name=vmsettings.ElementName, version=vm_version_str,
                min_version=min_version, max_version=max_version)

        return inner
    return wrapper



# Copyright 2017 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes

from oslo_log import log as logging

from os_win.utils import win32utils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi import libs as w_lib
from os_win.utils.winapi.libs import kernel32 as kernel32_struct

kernel32 = w_lib.get_shared_lib_handle(w_lib.KERNEL32)

LOG = logging.getLogger(__name__)


class ProcessUtils(object):
    def __init__(self):
        self._win32_utils = win32utils.Win32Utils()

    def _run_and_check_output(self, *args, **kwargs):
        kwargs.update(kernel32_lib_func=True)
        return self._win32_utils.run_and_check_output(*args, **kwargs)

    def create_job_object(self, name=None):
        """Create or open a job object.

        :param name: (Optional) the job name.
        :returns: a handle of the created job.
        """
        pname = None if name is None else ctypes.c_wchar_p(name)
        return self._run_and_check_output(kernel32.CreateJobObjectW,
                                          None,  # job security attributes
                                          pname,
                                          error_ret_vals=[None])

    def set_information_job_object(self, job_handle, job_object_info_class,
                                   job_object_info):
        self._run_and_check_output(kernel32.SetInformationJobObject,
                                   job_handle,
                                   job_object_info_class,
                                   ctypes.byref(job_object_info),
                                   ctypes.sizeof(job_object_info))

    def assign_process_to_job_object(self, job_handle, process_handle):
        self._run_and_check_output(kernel32.AssignProcessToJobObject,
                                   job_handle, process_handle)

    def open_process(self, pid, desired_access, inherit_handle=False):
        """Open an existing process."""
        return self._run_and_check_output(kernel32.OpenProcess,
                                          desired_access,
                                          inherit_handle,
                                          pid,
                                          error_ret_vals=[None])

    def kill_process_on_job_close(self, pid):
        """Associates a new job to the specified process.

        The process is immediately killed when the last job handle is closed.
        This mechanism can be useful when ensuring that child processes get
        killed along with a parent process.

        This method does not check if the specified process is already part of
        a job. Starting with WS 2012, nested jobs are available.

        :returns: the job handle, if a job was successfully created and
                  associated with the process, otherwise "None".
        """

        process_handle = None
        job_handle = None
        job_associated = False

        try:
            desired_process_access = (w_const.PROCESS_SET_QUOTA |
                                      w_const.PROCESS_TERMINATE)
            process_handle = self.open_process(pid, desired_process_access)
            job_handle = self.create_job_object()

            job_info = kernel32_struct.JOBOBJECT_EXTENDED_LIMIT_INFORMATION()
            job_info.BasicLimitInformation.LimitFlags = (
                w_const.JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE)
            job_info_class = w_const.JobObjectExtendedLimitInformation

            self.set_information_job_object(job_handle,
                                            job_info_class,
                                            job_info)

            self.assign_process_to_job_object(job_handle, process_handle)
            job_associated = True
        finally:
            if process_handle:
                self._win32_utils.close_handle(process_handle)

            if not job_associated and job_handle:
                # We have an unassociated job object. Closing the handle
                # will also destroy the job object.
                self._win32_utils.close_handle(job_handle)

        return job_handle

    def wait_for_multiple_processes(self, pids, wait_all=True,
                                    milliseconds=w_const.INFINITE):
        handles = []
        try:
            for pid in pids:
                handle = self.open_process(pid,
                                           desired_access=w_const.SYNCHRONIZE)
                handles.append(handle)

            return self._win32_utils.wait_for_multiple_objects(
                handles, wait_all, milliseconds)
        finally:
            for handle in handles:
                self._win32_utils.close_handle(handle)



# Copyright 2015 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base Utility class for operations on Hyper-V.
"""

import time

import monotonic
from oslo_log import log as logging

from os_win import _utils
import os_win.conf
from os_win import constants
from os_win import exceptions
from os_win.utils import baseutils

CONF = os_win.conf.CONF

LOG = logging.getLogger(__name__)


class JobUtils(baseutils.BaseUtilsVirt):

    _CONCRETE_JOB_CLASS = "Msvm_ConcreteJob"

    _KILL_JOB_STATE_CHANGE_REQUEST = 5

    _completed_job_states = [constants.JOB_STATE_COMPLETED,
                             constants.JOB_STATE_TERMINATED,
                             constants.JOB_STATE_KILLED,
                             constants.JOB_STATE_COMPLETED_WITH_WARNINGS,
                             constants.JOB_STATE_EXCEPTION]
    _successful_job_states = [constants.JOB_STATE_COMPLETED,
                              constants.JOB_STATE_COMPLETED_WITH_WARNINGS]

    def check_ret_val(self, ret_val, job_path, success_values=[0]):
        """Checks that the job represented by the given arguments succeeded.

        Some Hyper-V operations are not atomic, and will return a reference
        to a job. In this case, this method will wait for the job's
        completion.

        :param ret_val: integer, representing the return value of the job.
            if the value is WMI_JOB_STATUS_STARTED or WMI_JOB_STATE_RUNNING,
            a job_path cannot be None.
        :param job_path: string representing the WMI object path of a
            Hyper-V job.
        :param success_values: list of return values that can be considered
            successful. WMI_JOB_STATUS_STARTED and WMI_JOB_STATE_RUNNING
            values are ignored.
        :raises exceptions.WMIJobFailed: if the given ret_val is
            WMI_JOB_STATUS_STARTED or WMI_JOB_STATE_RUNNING and the state of
            job represented by the given job_path is not
            WMI_JOB_STATE_COMPLETED or JOB_STATE_COMPLETED_WITH_WARNINGS, or
            if the given ret_val is not in the list of given success_values.
        """
        if ret_val in [constants.WMI_JOB_STATUS_STARTED,
                       constants.WMI_JOB_STATE_RUNNING]:
            return self._wait_for_job(job_path)
        elif ret_val not in success_values:
            raise exceptions.WMIJobFailed(error_code=ret_val,
                                          job_state=None,
                                          error_summ_desc=None,
                                          error_desc=None)

    def _wait_for_job(self, job_path):
        """Poll WMI job state and wait for completion."""

        job_wmi_path = job_path.replace('\\', '/')
        job = self._get_wmi_obj(job_wmi_path)

        # We'll log the job status from time to time.
        last_report_time = 0
        report_interval = 5

        while not self._is_job_completed(job):
            now = monotonic.monotonic()
            if now - last_report_time > report_interval:
                job_details = self._get_job_details(job)
                LOG.debug("Waiting for WMI job: %s.", job_details)
                last_report_time = now

            time.sleep(0.1)
            job = self._get_wmi_obj(job_wmi_path)

        job_state = job.JobState
        err_code = job.ErrorCode

        # We'll raise an exception for killed jobs.
        job_failed = job_state not in self._successful_job_states or err_code
        job_warnings = job_state == constants.JOB_STATE_COMPLETED_WITH_WARNINGS
        job_details = self._get_job_details(
            job, extended=(job_failed or job_warnings))

        if job_failed:
            err_sum_desc = getattr(job, 'ErrorSummaryDescription', None)
            err_desc = job.ErrorDescription

            LOG.error("WMI job failed: %s.", job_details)
            raise exceptions.WMIJobFailed(job_state=job_state,
                                          error_code=err_code,
                                          error_summ_desc=err_sum_desc,
                                          error_desc=err_desc)

        if job_warnings:
            LOG.warning("WMI job completed with warnings. For detailed "
                        "information, please check the Windows event logs. "
                        "Job details: %s.", job_details)
        else:
            LOG.debug("WMI job succeeded: %s.", job_details)

        return job

    @staticmethod
    def _get_job_details(job, extended=False):
        basic_details = [
            "InstanceID", "Description", "ElementName", "JobStatus",
            "ElapsedTime", "Cancellable", "JobType", "Owner",
            "PercentComplete"]
        extended_details = [
            "JobState", "StatusDescriptions", "OperationalStatus",
            "TimeSubmitted", "UntilTime", "TimeOfLastStateChange",
            "DetailedStatus", "LocalOrUtcTime",
            "ErrorCode", "ErrorDescription", "ErrorSummaryDescription"]

        fields = list(basic_details)
        if extended:
            fields += extended_details

        details = {}
        for field in fields:
            try:
                details[field] = getattr(job, field)
            except AttributeError:
                continue

        return details

    def _get_pending_jobs_affecting_element(self, element):
        # Msvm_AffectedJobElement is in fact an association between
        # the affected element and the affecting job.
        mappings = self._conn.Msvm_AffectedJobElement(
            AffectedElement=element.path_())
        pending_jobs = []
        for mapping in mappings:
            try:
                if mapping.AffectingElement and not self._is_job_completed(
                        mapping.AffectingElement):
                    pending_jobs.append(mapping.AffectingElement)

            except exceptions.x_wmi as ex:
                # NOTE(claudiub): we can ignore "Not found" type exceptions.
                if not _utils._is_not_found_exc(ex):
                    raise

        return pending_jobs

    def _stop_jobs(self, element):
        pending_jobs = self._get_pending_jobs_affecting_element(element)
        for job in pending_jobs:
            job_details = self._get_job_details(job, extended=True)
            try:
                if not job.Cancellable:
                    LOG.debug("Got request to terminate "
                              "non-cancelable job: %s.", job_details)
                    continue

                job.RequestStateChange(
                    self._KILL_JOB_STATE_CHANGE_REQUEST)
            except exceptions.x_wmi as ex:
                # The job may had been completed right before we've
                # attempted to kill it.
                if not _utils._is_not_found_exc(ex):
                    LOG.debug("Failed to stop job. Exception: %s. "
                              "Job details: %s.", ex, job_details)

        pending_jobs = self._get_pending_jobs_affecting_element(element)
        if pending_jobs:
            pending_job_details = [self._get_job_details(job, extended=True)
                                   for job in pending_jobs]
            LOG.debug("Attempted to terminate jobs "
                      "affecting element %(element)s but "
                      "%(pending_count)s jobs are still pending: "
                      "%(pending_jobs)s.",
                      dict(element=element,
                           pending_count=len(pending_jobs),
                           pending_jobs=pending_job_details))
            raise exceptions.JobTerminateFailed()

    def _is_job_completed(self, job):
        return job.JobState in self._completed_job_states

    def stop_jobs(self, element, timeout=None):
        """Stops the Hyper-V jobs associated with the given resource.

        :param element: string representing the path of the Hyper-V resource
            whose jobs will be stopped.
        :param timeout: the maximum amount of time allowed to stop all the
            given resource's jobs.
        :raises exceptions.JobTerminateFailed: if there are still pending jobs
            associated with the given resource and the given timeout amount of
            time has passed.
        """
        if timeout is None:
            timeout = CONF.os_win.wmi_job_terminate_timeout

        @_utils.retry_decorator(exceptions=exceptions.JobTerminateFailed,
                                timeout=timeout, max_retry_count=None)
        def _stop_jobs_with_timeout():
            self._stop_jobs(element)

        _stop_jobs_with_timeout()

    @_utils.not_found_decorator()
    @_utils.retry_decorator(exceptions=exceptions.HyperVException)
    def add_virt_resource(self, virt_resource, parent):
        (job_path, new_resources,
         ret_val) = self._vs_man_svc.AddResourceSettings(
            parent.path_(), [virt_resource.GetText_(1)])
        self.check_ret_val(ret_val, job_path)
        return new_resources

    # modify_virt_resource can fail, especially while setting up the VM's
    # serial port connection. Retrying the operation will yield success.
    @_utils.not_found_decorator()
    @_utils.retry_decorator(exceptions=exceptions.HyperVException)
    def modify_virt_resource(self, virt_resource):
        (job_path, out_set_data,
         ret_val) = self._vs_man_svc.ModifyResourceSettings(
            ResourceSettings=[virt_resource.GetText_(1)])
        self.check_ret_val(ret_val, job_path)

    def remove_virt_resource(self, virt_resource):
        self.remove_multiple_virt_resources([virt_resource])

    @_utils.not_found_decorator()
    @_utils.retry_decorator(exceptions=exceptions.HyperVException)
    def remove_multiple_virt_resources(self, virt_resources):
        (job, ret_val) = self._vs_man_svc.RemoveResourceSettings(
            ResourceSettings=[r.path_() for r in virt_resources])
        self.check_ret_val(ret_val, job)

    def add_virt_feature(self, virt_feature, parent):
        self.add_multiple_virt_features([virt_feature], parent)

    @_utils.not_found_decorator()
    @_utils.retry_decorator(exceptions=exceptions.HyperVException)
    def add_multiple_virt_features(self, virt_features, parent):
        (job_path, out_set_data,
         ret_val) = self._vs_man_svc.AddFeatureSettings(
            parent.path_(), [f.GetText_(1) for f in virt_features])
        self.check_ret_val(ret_val, job_path)

    @_utils.not_found_decorator()
    @_utils.retry_decorator(exceptions=exceptions.HyperVException)
    def modify_virt_feature(self, virt_feature):
        (job_path, out_set_data,
         ret_val) = self._vs_man_svc.ModifyFeatureSettings(
            FeatureSettings=[virt_feature.GetText_(1)])
        self.check_ret_val(ret_val, job_path)

    def remove_virt_feature(self, virt_feature):
        self.remove_multiple_virt_features([virt_feature])

    @_utils.not_found_decorator()
    def remove_multiple_virt_features(self, virt_features):
        (job_path, ret_val) = self._vs_man_svc.RemoveFeatureSettings(
            FeatureSettings=[f.path_() for f in virt_features])
        self.check_ret_val(ret_val, job_path)



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils import _wqlutils
from os_win.utils.compute import vmutils
from os_win.utils.metrics import metricsutils
from os_win import utilsfactory


class MetricsUtilsTestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the Hyper-V MetricsUtils class."""

    _FAKE_RET_VAL = 0
    _FAKE_PORT = "fake's port name"

    _autospec_classes = [
        vmutils.VMUtils,
    ]

    def setUp(self):
        super(MetricsUtilsTestCase, self).setUp()

        mock.patch.object(utilsfactory, 'get_vmutils',
                          mock.Mock(return_value=vmutils.VMUtils)).start()

        self.utils = metricsutils.MetricsUtils()
        self.utils._conn_attr = mock.MagicMock()

    def test_cache_metrics_defs(self):
        mock_metric_def = mock.Mock(ElementName=mock.sentinel.elementname)
        self.utils._conn.CIM_BaseMetricDefinition.return_value = [
            mock_metric_def]
        self.utils._cache_metrics_defs()
        expected_cache_metrics = {mock.sentinel.elementname: mock_metric_def}
        self.assertEqual(expected_cache_metrics, self.utils._metrics_defs_obj)

    @mock.patch.object(metricsutils.MetricsUtils, '_enable_metrics')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm_resources')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm')
    def test_enable_vm_metrics_collection(
            self, mock_get_vm, mock_get_vm_resources, mock_enable_metrics):
        mock_vm = mock_get_vm.return_value
        mock_disk = mock.MagicMock()
        mock_dvd = mock.MagicMock(
            ResourceSubType=self.utils._DVD_DISK_RES_SUB_TYPE)
        mock_get_vm_resources.return_value = [mock_disk, mock_dvd]

        self.utils.enable_vm_metrics_collection(mock.sentinel.vm_name)

        metrics_names = [self.utils._CPU_METRICS,
                         self.utils._MEMORY_METRICS]
        mock_enable_metrics.assert_has_calls(
            [mock.call(mock_disk), mock.call(mock_vm, metrics_names)])

    @mock.patch.object(metricsutils.MetricsUtils, '_enable_metrics')
    def test_enable_disk_metrics_collection(self, mock_enable_metrics):
        mock_get_disk = (
            self.utils._vmutils._get_mounted_disk_resource_from_path)

        self.utils.enable_disk_metrics_collection(
            mock.sentinel.disk_path,
            mock.sentinel.is_physical,
            mock.sentinel.serial)

        mock_get_disk.assert_called_once_with(
            mock.sentinel.disk_path,
            is_physical=mock.sentinel.is_physical,
            serial=mock.sentinel.serial)
        mock_enable_metrics.assert_called_once_with(mock_get_disk.return_value)

    @mock.patch.object(metricsutils.MetricsUtils, '_enable_metrics')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_switch_port')
    def test_enable_switch_port_metrics_collection(self, mock_get_port,
                                                   mock_enable_metrics):
        self.utils.enable_port_metrics_collection(mock.sentinel.port_name)

        mock_get_port.assert_called_once_with(mock.sentinel.port_name)
        metrics = [self.utils._NET_IN_METRICS,
                   self.utils._NET_OUT_METRICS]
        mock_enable_metrics.assert_called_once_with(
            mock_get_port.return_value, metrics)

    def _check_enable_metrics(self, metrics=None, definition=None):
        mock_element = mock.MagicMock()
        self.utils._metrics_svc.ControlMetrics.return_value = [0]

        self.utils._enable_metrics(mock_element, metrics)

        self.utils._metrics_svc.ControlMetrics.assert_called_once_with(
            Subject=mock_element.path_.return_value,
            Definition=definition,
            MetricCollectionEnabled=self.utils._METRICS_ENABLED)

    def test_enable_metrics_no_metrics(self):
        self._check_enable_metrics()

    def test_enable_metrics(self):
        metrics_name = self.utils._CPU_METRICS
        metrics_def = mock.MagicMock()
        self.utils._metrics_defs_obj = {metrics_name: metrics_def}
        self._check_enable_metrics([metrics_name, mock.sentinel.metrics_name],
                                   metrics_def.path_.return_value)

    def test_enable_metrics_exception(self):
        metric_name = self.utils._CPU_METRICS
        metric_def = mock.MagicMock()
        self.utils._metrics_defs_obj = {metric_name: metric_def}

        self.utils._metrics_svc.ControlMetrics.return_value = [1]
        self.assertRaises(exceptions.OSWinException,
                          self.utils._enable_metrics,
                          mock.MagicMock(),
                          [metric_name])

    @mock.patch.object(metricsutils.MetricsUtils, '_get_metrics')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm_resources')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm')
    def test_get_cpu_metrics(self, mock_get_vm, mock_get_vm_resources,
                             mock_get_metrics):
        fake_cpu_count = 2
        fake_uptime = 1000
        fake_cpu_metrics_val = 2000

        self.utils._metrics_defs_obj = {
            self.utils._CPU_METRICS: mock.sentinel.metrics}

        mock_vm = mock_get_vm.return_value
        mock_vm.OnTimeInMilliseconds = fake_uptime
        mock_cpu = mock.MagicMock(VirtualQuantity=fake_cpu_count)
        mock_get_vm_resources.return_value = [mock_cpu]

        mock_metric = mock.MagicMock(MetricValue=fake_cpu_metrics_val)
        mock_get_metrics.return_value = [mock_metric]

        cpu_metrics = self.utils.get_cpu_metrics(mock.sentinel.vm_name)

        self.assertEqual(3, len(cpu_metrics))
        self.assertEqual(fake_cpu_metrics_val, cpu_metrics[0])
        self.assertEqual(fake_cpu_count, cpu_metrics[1])
        self.assertEqual(fake_uptime, cpu_metrics[2])

        mock_get_vm.assert_called_once_with(mock.sentinel.vm_name)
        mock_get_vm_resources.assert_called_once_with(
            mock.sentinel.vm_name, self.utils._PROCESSOR_SETTING_DATA_CLASS)
        mock_get_metrics.assert_called_once_with(mock_vm,
                                                 mock.sentinel.metrics)

    @mock.patch.object(metricsutils.MetricsUtils, '_get_metrics')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm')
    def test_get_memory_metrics(self, mock_get_vm, mock_get_metrics):
        mock_vm = mock_get_vm.return_value
        self.utils._metrics_defs_obj = {
            self.utils._MEMORY_METRICS: mock.sentinel.metrics}

        metrics_memory = mock.MagicMock()
        metrics_memory.MetricValue = 3
        mock_get_metrics.return_value = [metrics_memory]

        response = self.utils.get_memory_metrics(mock.sentinel.vm_name)

        self.assertEqual(3, response)
        mock_get_vm.assert_called_once_with(mock.sentinel.vm_name)
        mock_get_metrics.assert_called_once_with(mock_vm,
                                                 mock.sentinel.metrics)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    @mock.patch.object(metricsutils.MetricsUtils,
                       '_sum_metrics_values_by_defs')
    @mock.patch.object(metricsutils.MetricsUtils,
                       '_get_metrics_value_instances')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm_resources')
    def test_get_vnic_metrics(self, mock_get_vm_resources,
                              mock_get_value_instances, mock_sum_by_defs,
                              mock_get_element_associated_class):
        fake_rx_mb = 1000
        fake_tx_mb = 2000

        self.utils._metrics_defs_obj = {
            self.utils._NET_IN_METRICS: mock.sentinel.net_in_metrics,
            self.utils._NET_OUT_METRICS: mock.sentinel.net_out_metrics}

        mock_port = mock.MagicMock(Parent=mock.sentinel.vnic_path)
        mock_vnic = mock.MagicMock(ElementName=mock.sentinel.element_name,
                                   Address=mock.sentinel.address)
        mock_vnic.path_.return_value = mock.sentinel.vnic_path
        mock_get_vm_resources.side_effect = [[mock_port], [mock_vnic]]
        mock_sum_by_defs.return_value = [fake_rx_mb, fake_tx_mb]

        vnic_metrics = list(
            self.utils.get_vnic_metrics(mock.sentinel.vm_name))

        self.assertEqual(1, len(vnic_metrics))
        self.assertEqual(fake_rx_mb, vnic_metrics[0]['rx_mb'])
        self.assertEqual(fake_tx_mb, vnic_metrics[0]['tx_mb'])
        self.assertEqual(mock.sentinel.element_name,
                         vnic_metrics[0]['element_name'])
        self.assertEqual(mock.sentinel.address, vnic_metrics[0]['address'])

        mock_get_vm_resources.assert_has_calls([
            mock.call(mock.sentinel.vm_name, self.utils._PORT_ALLOC_SET_DATA),
            mock.call(mock.sentinel.vm_name,
                      self.utils._SYNTH_ETH_PORT_SET_DATA)])
        mock_get_value_instances.assert_called_once_with(
            mock_get_element_associated_class.return_value,
            self.utils._BASE_METRICS_VALUE)
        mock_sum_by_defs.assert_called_once_with(
            mock_get_value_instances.return_value,
            [mock.sentinel.net_in_metrics, mock.sentinel.net_out_metrics])

    @mock.patch.object(metricsutils.MetricsUtils, '_get_metrics_values')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm_resources')
    def test_get_disk_metrics(self, mock_get_vm_resources,
                              mock_get_metrics_values):
        fake_read_mb = 1000
        fake_write_mb = 2000

        self.utils._metrics_defs_obj = {
            self.utils._DISK_RD_METRICS: mock.sentinel.disk_rd_metrics,
            self.utils._DISK_WR_METRICS: mock.sentinel.disk_wr_metrics}

        mock_disk = mock.MagicMock(HostResource=[mock.sentinel.host_resource],
                                   InstanceID=mock.sentinel.instance_id)
        mock_get_vm_resources.return_value = [mock_disk]
        mock_get_metrics_values.return_value = [fake_read_mb, fake_write_mb]

        disk_metrics = list(
            self.utils.get_disk_metrics(mock.sentinel.vm_name))

        self.assertEqual(1, len(disk_metrics))
        self.assertEqual(fake_read_mb, disk_metrics[0]['read_mb'])
        self.assertEqual(fake_write_mb, disk_metrics[0]['write_mb'])
        self.assertEqual(mock.sentinel.instance_id,
                         disk_metrics[0]['instance_id'])
        self.assertEqual(mock.sentinel.host_resource,
                         disk_metrics[0]['host_resource'])

        mock_get_vm_resources.assert_called_once_with(
            mock.sentinel.vm_name,
            self.utils._STORAGE_ALLOC_SETTING_DATA_CLASS)
        metrics = [mock.sentinel.disk_rd_metrics,
                   mock.sentinel.disk_wr_metrics]
        mock_get_metrics_values.assert_called_once_with(mock_disk, metrics)

    @mock.patch.object(metricsutils.MetricsUtils, '_get_metrics_values')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm_resources')
    def test_get_disk_latency_metrics(self, mock_get_vm_resources,
                                      mock_get_metrics_values):
        self.utils._metrics_defs_obj = {
            self.utils._DISK_LATENCY_METRICS: mock.sentinel.metrics}

        mock_disk = mock.MagicMock(HostResource=[mock.sentinel.host_resource],
                                   InstanceID=mock.sentinel.instance_id)
        mock_get_vm_resources.return_value = [mock_disk]
        mock_get_metrics_values.return_value = [mock.sentinel.latency]

        disk_metrics = list(
            self.utils.get_disk_latency_metrics(mock.sentinel.vm_name))

        self.assertEqual(1, len(disk_metrics))
        self.assertEqual(mock.sentinel.latency,
                         disk_metrics[0]['disk_latency'])
        self.assertEqual(mock.sentinel.instance_id,
                         disk_metrics[0]['instance_id'])
        mock_get_vm_resources.assert_called_once_with(
            mock.sentinel.vm_name,
            self.utils._STORAGE_ALLOC_SETTING_DATA_CLASS)
        mock_get_metrics_values.assert_called_once_with(
            mock_disk, [mock.sentinel.metrics])

    @mock.patch.object(metricsutils.MetricsUtils, '_get_metrics_values')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm_resources')
    def test_get_disk_iops_metrics(self, mock_get_vm_resources,
                                   mock_get_metrics_values):
        self.utils._metrics_defs_obj = {
            self.utils._DISK_IOPS_METRICS: mock.sentinel.metrics}
        mock_disk = mock.MagicMock(HostResource=[mock.sentinel.host_resource],
                                   InstanceID=mock.sentinel.instance_id)
        mock_get_vm_resources.return_value = [mock_disk]
        mock_get_metrics_values.return_value = [mock.sentinel.iops]

        disk_metrics = list(
            self.utils.get_disk_iops_count(mock.sentinel.vm_name))

        self.assertEqual(1, len(disk_metrics))
        self.assertEqual(mock.sentinel.iops,
                         disk_metrics[0]['iops_count'])
        self.assertEqual(mock.sentinel.instance_id,
                         disk_metrics[0]['instance_id'])
        mock_get_vm_resources.assert_called_once_with(
            mock.sentinel.vm_name,
            self.utils._STORAGE_ALLOC_SETTING_DATA_CLASS)
        mock_get_metrics_values.assert_called_once_with(
            mock_disk, [mock.sentinel.metrics])

    def test_sum_metrics_values(self):
        mock_metric = mock.MagicMock(MetricValue='100')
        result = self.utils._sum_metrics_values([mock_metric] * 2)
        self.assertEqual(200, result)

    def test_sum_metrics_values_by_defs(self):
        mock_metric = mock.MagicMock(MetricDefinitionId=mock.sentinel.def_id,
                                     MetricValue='100')
        mock_metric_useless = mock.MagicMock(MetricValue='200')
        mock_metric_def = mock.MagicMock(Id=mock.sentinel.def_id)

        result = self.utils._sum_metrics_values_by_defs(
            [mock_metric, mock_metric_useless], [None, mock_metric_def])

        self.assertEqual([0, 100], result)

    def test_get_metrics_value_instances(self):
        FAKE_CLASS_NAME = "FAKE_CLASS"
        mock_el_metric = mock.MagicMock()
        mock_el_metric_2 = mock.MagicMock()
        mock_el_metric_2.path.return_value = mock.Mock(Class=FAKE_CLASS_NAME)

        self.utils._conn.Msvm_MetricForME.side_effect = [
            [], [mock.Mock(Dependent=mock_el_metric_2)]]

        returned = self.utils._get_metrics_value_instances(
            [mock_el_metric, mock_el_metric_2], FAKE_CLASS_NAME)

        expected_return = [mock_el_metric_2]
        self.assertEqual(expected_return, returned)

    @mock.patch.object(metricsutils.MetricsUtils,
                       '_sum_metrics_values_by_defs')
    def test_get_metrics_values(self, mock_sum_by_defs):
        mock_element = mock.MagicMock()
        self.utils._conn.Msvm_MetricForME.return_value = [
            mock.Mock(Dependent=mock.sentinel.metric),
            mock.Mock(Dependent=mock.sentinel.another_metric)]

        resulted_metrics_sum = self.utils._get_metrics_values(
            mock_element, mock.sentinel.metrics_defs)

        self.utils._conn.Msvm_MetricForME.assert_called_once_with(
            Antecedent=mock_element.path_.return_value)
        mock_sum_by_defs.assert_called_once_with(
            [mock.sentinel.metric, mock.sentinel.another_metric],
            mock.sentinel.metrics_defs)
        expected_metrics_sum = mock_sum_by_defs.return_value
        self.assertEqual(expected_metrics_sum, resulted_metrics_sum)

    @mock.patch.object(metricsutils.MetricsUtils, '_filter_metrics')
    def test_get_metrics(self, mock_filter_metrics):
        mock_metric = mock.MagicMock()
        mock_element = mock.MagicMock()
        self.utils._conn.Msvm_MetricForME.return_value = [mock_metric]

        result = self.utils._get_metrics(mock_element,
                                         mock.sentinel.metrics_def)

        self.assertEqual(mock_filter_metrics.return_value, result)
        self.utils._conn.Msvm_MetricForME.assert_called_once_with(
            Antecedent=mock_element.path_.return_value)
        mock_filter_metrics.assert_called_once_with(
            [mock_metric.Dependent],
            mock.sentinel.metrics_def)

    def test_filter_metrics(self):
        mock_metric = mock.MagicMock(MetricDefinitionId=mock.sentinel.def_id)
        mock_bad_metric = mock.MagicMock()
        mock_metric_def = mock.MagicMock(Id=mock.sentinel.def_id)

        result = self.utils._filter_metrics([mock_bad_metric, mock_metric],
                                            mock_metric_def)

        self.assertEqual([mock_metric], result)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    @mock.patch.object(metricsutils.MetricsUtils, '_get_vm_setting_data')
    def test_get_vm_resources(self, mock_get_vm_setting_data,
                              mock_get_element_associated_class):
        result = self.utils._get_vm_resources(mock.sentinel.vm_name,
                                              mock.sentinel.resource_class)

        mock_get_vm_setting_data.assert_called_once_with(mock.sentinel.vm_name)
        vm_setting_data = mock_get_vm_setting_data.return_value
        mock_get_element_associated_class.assert_called_once_with(
            self.utils._conn, mock.sentinel.resource_class,
            element_instance_id=vm_setting_data.InstanceID)
        self.assertEqual(mock_get_element_associated_class.return_value,
                         result)

    @mock.patch.object(metricsutils.MetricsUtils, '_unique_result')
    def test_get_vm(self, mock_unique_result):
        result = self.utils._get_vm(mock.sentinel.vm_name)

        self.assertEqual(mock_unique_result.return_value, result)
        conn_class = self.utils._conn.Msvm_ComputerSystem
        conn_class.assert_called_once_with(ElementName=mock.sentinel.vm_name)
        mock_unique_result.assert_called_once_with(conn_class.return_value,
                                                   mock.sentinel.vm_name)

    @mock.patch.object(metricsutils.MetricsUtils, '_unique_result')
    def test_get_switch_port(self, mock_unique_result):
        result = self.utils._get_switch_port(mock.sentinel.port_name)

        self.assertEqual(mock_unique_result.return_value, result)
        conn_class = self.utils._conn.Msvm_EthernetPortAllocationSettingData
        conn_class.assert_called_once_with(ElementName=mock.sentinel.port_name)
        mock_unique_result.assert_called_once_with(conn_class.return_value,
                                                   mock.sentinel.port_name)

    @mock.patch.object(metricsutils.MetricsUtils, '_unique_result')
    def test_get_vm_setting_data(self, mock_unique_result):
        result = self.utils._get_vm_setting_data(mock.sentinel.vm_name)

        self.assertEqual(mock_unique_result.return_value, result)
        conn_class = self.utils._conn.Msvm_VirtualSystemSettingData
        conn_class.assert_called_once_with(ElementName=mock.sentinel.vm_name)
        mock_unique_result.assert_called_once_with(conn_class.return_value,
                                                   mock.sentinel.vm_name)

    def test_unique_result_not_found(self):
        self.assertRaises(exceptions.NotFound,
                          self.utils._unique_result,
                          [], mock.sentinel.resource_name)

    def test_unique_result_duplicate(self):
        self.assertRaises(exceptions.OSWinException,
                          self.utils._unique_result,
                          [mock.ANY, mock.ANY], mock.sentinel.resource_name)

    def test_unique_result(self):
        result = self.utils._unique_result([mock.sentinel.obj],
                                           mock.sentinel.resource_name)
        self.assertEqual(mock.sentinel.obj, result)



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import six

from os_win._i18n import _
from os_win import constants
from os_win import exceptions
from os_win.utils import baseutils
from os_win.utils import hostutils
from os_win.utils import pathutils
from os_win.utils import win32utils

LOG = logging.getLogger(__name__)


@six.add_metaclass(baseutils.SynchronizedMeta)
class ISCSITargetUtils(baseutils.BaseUtils):
    ID_METHOD_DNS_NAME = 1
    ID_METHOD_IPV4_ADDR = 2
    ID_METHOD_MAC_ADDR = 3
    ID_METHOD_IQN = 4
    ID_METHOD_IPV6_ADDR = 5

    _ERR_FILE_EXISTS = 80

    def __init__(self):
        self._conn_wmi = self._get_wmi_conn('//./root/wmi')
        self._ensure_wt_provider_available()

        self._pathutils = pathutils.PathUtils()
        self._hostutils = hostutils.HostUtils()
        self._win32utils = win32utils.Win32Utils()

        self._win_gteq_6_3 = self._hostutils.check_min_windows_version(6, 3)

    def _ensure_wt_provider_available(self):
        try:
            self._conn_wmi.WT_Portal
        except AttributeError:
            err_msg = _("The Windows iSCSI target provider is not available.")
            raise exceptions.ISCSITargetException(err_msg)

    def get_supported_disk_format(self):
        return (constants.DISK_FORMAT_VHDX
                if self._win_gteq_6_3 else constants.DISK_FORMAT_VHD)

    def get_supported_vhd_type(self):
        return (constants.VHD_TYPE_DYNAMIC
                if self._win_gteq_6_3 else constants.VHD_TYPE_FIXED)

    def get_portal_locations(self, available_only=True,
                             fail_if_none_found=True):
        wt_portals = self._conn_wmi.WT_Portal()

        if available_only:
            wt_portals = list(filter(lambda portal: portal.Listen, wt_portals))

        if not wt_portals and fail_if_none_found:
            err_msg = _("No valid iSCSI portal was found.")
            raise exceptions.ISCSITargetException(err_msg)

        portal_locations = [self._get_portal_location(portal)
                            for portal in wt_portals]
        return portal_locations

    def _get_portal_location(self, wt_portal):
        return '%s:%s' % (wt_portal.Address, wt_portal.Port)

    def _get_wt_host(self, target_name, fail_if_not_found=True):
        hosts = self._conn_wmi.WT_Host(HostName=target_name)

        if hosts:
            return hosts[0]
        elif fail_if_not_found:
            err_msg = _('Could not find iSCSI target %s')
            raise exceptions.ISCSITargetException(err_msg % target_name)

    def _get_wt_disk(self, description, fail_if_not_found=True):
        # We can retrieve WT Disks only by description.
        wt_disks = self._conn_wmi.WT_Disk(Description=description)
        if wt_disks:
            return wt_disks[0]
        elif fail_if_not_found:
            err_msg = _('Could not find WT Disk: %s')
            raise exceptions.ISCSITargetException(err_msg % description)

    def _get_wt_snapshot(self, description, fail_if_not_found=True):
        wt_snapshots = self._conn_wmi.WT_Snapshot(Description=description)
        if wt_snapshots:
            return wt_snapshots[0]
        elif fail_if_not_found:
            err_msg = _('Could not find WT Snapshot: %s')
            raise exceptions.ISCSITargetException(err_msg % description)

    def _get_wt_idmethod(self, initiator, target_name):
        wt_idmethod = self._conn_wmi.WT_IDMethod(HostName=target_name,
                                                 Value=initiator)
        if wt_idmethod:
            return wt_idmethod[0]

    def create_iscsi_target(self, target_name, fail_if_exists=False):
        """Creates ISCSI target."""
        try:
            self._conn_wmi.WT_Host.NewHost(HostName=target_name)
        except exceptions.x_wmi as wmi_exc:
            err_code = self._win32utils.get_com_err_code(wmi_exc.com_error)
            target_exists = err_code == self._ERR_FILE_EXISTS

            if not target_exists or fail_if_exists:
                err_msg = _('Failed to create iSCSI target: %s.')
                raise exceptions.ISCSITargetWMIException(err_msg % target_name,
                                                         wmi_exc=wmi_exc)
            else:
                LOG.info('The iSCSI target %s already exists.',
                         target_name)

    def delete_iscsi_target(self, target_name):
        """Removes ISCSI target."""
        try:
            wt_host = self._get_wt_host(target_name, fail_if_not_found=False)
            if not wt_host:
                LOG.debug('Skipping deleting target %s as it does not '
                          'exist.', target_name)
                return
            wt_host.RemoveAllWTDisks()
            wt_host.Delete_()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _("Failed to delete ISCSI target %s")
            raise exceptions.ISCSITargetWMIException(err_msg % target_name,
                                                     wmi_exc=wmi_exc)

    def iscsi_target_exists(self, target_name):
        wt_host = self._get_wt_host(target_name, fail_if_not_found=False)
        return wt_host is not None

    def get_target_information(self, target_name):
        wt_host = self._get_wt_host(target_name)

        info = {}
        info['target_iqn'] = wt_host.TargetIQN
        info['enabled'] = wt_host.Enabled
        info['connected'] = bool(wt_host.Status)

        # Note(lpetrut): Cinder uses only one-way CHAP authentication.
        if wt_host.EnableCHAP:
            info['auth_method'] = 'CHAP'
            info['auth_username'] = wt_host.CHAPUserName
            info['auth_password'] = wt_host.CHAPSecret

        return info

    def set_chap_credentials(self, target_name, chap_username, chap_password):
        try:
            wt_host = self._get_wt_host(target_name)
            wt_host.EnableCHAP = True
            wt_host.CHAPUserName = chap_username
            wt_host.CHAPSecret = chap_password
            wt_host.put()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Failed to set CHAP credentials on target %s.')
            raise exceptions.ISCSITargetWMIException(err_msg % target_name,
                                                     wmi_exc=wmi_exc)

    def associate_initiator_with_iscsi_target(self, initiator,
                                              target_name,
                                              id_method=ID_METHOD_IQN):
        wt_idmethod = self._get_wt_idmethod(initiator, target_name)
        if wt_idmethod:
            return

        try:
            wt_idmethod = self._conn_wmi.WT_IDMethod.new()
            wt_idmethod.HostName = target_name
            wt_idmethod.Method = id_method
            wt_idmethod.Value = initiator
            wt_idmethod.put()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Could not associate initiator %(initiator)s to '
                        'iSCSI target: %(target_name)s.')
            raise exceptions.ISCSITargetWMIException(
                err_msg % dict(initiator=initiator,
                               target_name=target_name),
                wmi_exc=wmi_exc)

    def deassociate_initiator(self, initiator, target_name):
        try:
            wt_idmethod = self._get_wt_idmethod(initiator, target_name)
            if wt_idmethod:
                wt_idmethod.Delete_()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Could not deassociate initiator %(initiator)s from '
                        'iSCSI target: %(target_name)s.')
            raise exceptions.ISCSITargetWMIException(
                err_msg % dict(initiator=initiator,
                               target_name=target_name),
                wmi_exc=wmi_exc)

    def create_wt_disk(self, vhd_path, wtd_name, size_mb=None):
        try:
            self._conn_wmi.WT_Disk.NewWTDisk(DevicePath=vhd_path,
                                             Description=wtd_name,
                                             SizeInMB=size_mb)
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Failed to create WT Disk. '
                        'VHD path: %(vhd_path)s '
                        'WT disk name: %(wtd_name)s')
            raise exceptions.ISCSITargetWMIException(
                err_msg % dict(vhd_path=vhd_path,
                               wtd_name=wtd_name),
                wmi_exc=wmi_exc)

    def import_wt_disk(self, vhd_path, wtd_name):
        """Import a vhd/x image to be used by Windows iSCSI targets."""
        try:
            self._conn_wmi.WT_Disk.ImportWTDisk(DevicePath=vhd_path,
                                                Description=wtd_name)
        except exceptions.x_wmi as wmi_exc:
            err_msg = _("Failed to import WT disk: %s.")
            raise exceptions.ISCSITargetWMIException(err_msg % vhd_path,
                                                     wmi_exc=wmi_exc)

    def change_wt_disk_status(self, wtd_name, enabled):
        try:
            wt_disk = self._get_wt_disk(wtd_name)
            wt_disk.Enabled = enabled
            wt_disk.put()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Could not change disk status. WT Disk name: %s')
            raise exceptions.ISCSITargetWMIException(err_msg % wtd_name,
                                                     wmi_exc=wmi_exc)

    def remove_wt_disk(self, wtd_name):
        try:
            wt_disk = self._get_wt_disk(wtd_name, fail_if_not_found=False)
            if wt_disk:
                wt_disk.Delete_()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _("Failed to remove WT disk: %s.")
            raise exceptions.ISCSITargetWMIException(err_msg % wtd_name,
                                                     wmi_exc=wmi_exc)

    def extend_wt_disk(self, wtd_name, additional_mb):
        try:
            wt_disk = self._get_wt_disk(wtd_name)
            wt_disk.Extend(additional_mb)
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Could not extend WT Disk %(wtd_name)s '
                        'with additional %(additional_mb)s MB.')
            raise exceptions.ISCSITargetWMIException(
                err_msg % dict(wtd_name=wtd_name,
                               additional_mb=additional_mb),
                wmi_exc=wmi_exc)

    def add_disk_to_target(self, wtd_name, target_name):
        """Adds the disk to the target."""
        try:
            wt_disk = self._get_wt_disk(wtd_name)
            wt_host = self._get_wt_host(target_name)
            wt_host.AddWTDisk(wt_disk.WTD)
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Could not add WTD Disk %(wtd_name)s to '
                        'iSCSI target %(target_name)s.')
            raise exceptions.ISCSITargetWMIException(
                err_msg % dict(wtd_name=wtd_name,
                               target_name=target_name),
                wmi_exc=wmi_exc)

    def create_snapshot(self, wtd_name, snapshot_name):
        """Driver entry point for creating a snapshot."""
        try:
            wt_disk = self._get_wt_disk(wtd_name)
            snap_id = self._conn_wmi.WT_Snapshot.Create(WTD=wt_disk.WTD)[0]

            wt_snap = self._conn_wmi.WT_Snapshot(Id=snap_id)[0]
            wt_snap.Description = snapshot_name
            wt_snap.put()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Failed to create snapshot. '
                        'WT Disk name: %(wtd_name)s '
                        'Snapshot name: %(snapshot_name)s')
            raise exceptions.ISCSITargetWMIException(
                err_msg % dict(wtd_name=wtd_name,
                               snapshot_name=snapshot_name),
                wmi_exc=wmi_exc)

    def export_snapshot(self, snapshot_name, dest_path):
        """Driver entry point for exporting snapshots as volumes."""
        try:
            wt_snap = self._get_wt_snapshot(snapshot_name)
            wt_disk_id = wt_snap.Export()[0]
            # This export is a read-only shadow copy, needing to be copied
            # to another disk.
            wt_disk = self._conn_wmi.WT_Disk(WTD=wt_disk_id)[0]
            wt_disk.Description = '%s-%s-temp' % (snapshot_name, wt_disk_id)
            wt_disk.put()
            src_path = wt_disk.DevicePath

            self._pathutils.copy(src_path, dest_path)

            wt_disk.Delete_()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Failed to export snapshot %(snapshot_name)s '
                        'to %(dest_path)s.')
            raise exceptions.ISCSITargetWMIException(
                err_msg % dict(snapshot_name=snapshot_name,
                               dest_path=dest_path),
                wmi_exc=wmi_exc)

    def delete_snapshot(self, snapshot_name):
        """Driver entry point for deleting a snapshot."""
        try:
            wt_snapshot = self._get_wt_snapshot(snapshot_name,
                                                fail_if_not_found=False)
            if wt_snapshot:
                wt_snapshot.Delete_()
        except exceptions.x_wmi as wmi_exc:
            err_msg = _('Failed delete snapshot %s.')
            raise exceptions.ISCSITargetWMIException(err_msg % snapshot_name,
                                                     wmi_exc=wmi_exc)
